#!/usr/bin/python3

import asyncio
"""
	Installer le module websockets via
	le dépôt  https://pypi.python.org/pypi/websockets
	ou la commande pip ->  pip install websockets
"""
import websockets
import json
import ssl
import time
import signal
import uuid
import sqlite3
import sys

def signal_handler(signal, frame):
	print('You pressed Ctrl+C!')
	if conn : conn.close()
	sys.exit(0)

def UID():
	return uuid.uuid1().urn.replace('urn:uuid:','')

@asyncio.coroutine
def Loop(websocket, path):
	id = UID()#Id Thread
	Users[id] = 'Anonymous'
	bdd.execute("INSERT INTO Users(uid, name) VALUES(?,?)", (id, 'Anonymous'))
	conn.commit()
	lastreadMsg[id] = []
	lastreadEvt[id] = []
	print(">Client Connection from {}, id {}".format(websocket.host, id))
	future = asyncio.Future()
	asyncio.async(Reception(future, websocket, id))
	yield from checkEvt(websocket, id, True)
	while True:
		if not websocket.open:
			del lastreadMsg[id]
			del lastreadEvt[id]
			del Users[id]
			print(">Client Disconnected")
			return
		try:
			future.result()
			del lastreadMsg[id]
			del lastreadEvt[id]
			del Users[id]
			print(">Client Disconnected")
			return
		except:
			{}
		yield from producer(websocket, id)
		yield from checkEvt(websocket, id)
		yield from asyncio.sleep(1)
	

@asyncio.coroutine
def Reception(future, websocket, id):
	while True:
		out = ""
		brut = yield from websocket.recv()
		if brut is None:
			future.set_result('Future is done!')
			return
		try:
			data = json.loads(brut)
		except ValueError as e:
			print("Client data was not a valid JSON text")
		if not isinstance(data, dict):return;
		print("<<<")
		if "type" in data and "data" in data  and "date" in data  and "init" in data :
			print("date : {}".format(data['date']))
			print("type : {}".format(data['type']))
			print("data : {}".format(data['data']))
			if data['type'] == "message":
				bdd.execute("INSERT INTO Messages(id, user, content, time) VALUES(?,?,?,?)", (UID(), id, data['data'], time.time()))
				conn.commit()
			elif data['type'] == "changeName":
				bdd.execute("UPDATE Users SET name=? WHERE uid=?", (data['data'], id))
				bdd.execute("INSERT INTO Events(eid,etype,init,msg,time) VALUES(?,?,?,?,?)", (UID(), "changeName", data['init'], "{} -> {}".format(Users[id],data['data']), time.time()))
				conn.commit()
				Users[id] = data['data']
			else:
				out = '{"type":"error", "code":"2"}'
		else : 
			print(data)
			out = '{"type":"error", "code":"1"}'
		if out != "":
			print(">>>{}".format(out))
			yield from websocket.send(out)

def producer(websocket, id):
	msgTable = []
	for row in bdd.execute('SELECT m.id,u.name,m.content,m.time FROM Messages as m JOIN Users as u ON(m.user = u.uid) ORDER BY time DESC LIMIT 5'):
		msgTable.insert(0, row)

	for line in msgTable:
		if line[0] not in lastreadMsg[id] :
			print(line)
			data = {'type':"message", 'user':line[1], 'time':line[3], 'id':line[0], 'data':line[2]}
			yield from websocket.send(json.dumps(data))
			lastreadMsg[id].insert(0, line[0])

	while len(lastreadMsg[id]) > 5 :
		lastreadMsg[id].pop()

def checkEvt(websocket, id, init = False):
	evtTable = []
	for row in bdd.execute('SELECT eid,etype,init,msg,time FROM Events ORDER BY time DESC LIMIT 5'):
		evtTable.insert(0, row)

	for line in evtTable:
		if line[0] not in lastreadEvt[id] :
			if line[2] == "false":
				print(line)
				data = {'type':"event", 'etype':line[1], 'msg':line[3], 'id':line[0]}
				if(init is False): yield from websocket.send(json.dumps(data))
			lastreadEvt[id].insert(0, line[0])

	while len(lastreadEvt[id]) > 5 :
		lastreadEvt[id].pop()


signal.signal(signal.SIGINT, signal_handler)
		
sslcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
sslcontext.load_cert_chain("ssl2.crt", "ssl2.key")

conn = sqlite3.connect('work.db')
bdd = conn.cursor()
bdd.execute('''CREATE TABLE IF NOT EXISTS 'Messages' ('id' TEXT PRIMARY KEY NOT NULL, 'user' TEXT NOT NULL, 'content' TEXT NOT NULL, 'time' INTEGER NOT NULL)''')
bdd.execute('''CREATE TABLE IF NOT EXISTS 'Users' ('uid' TEXT PRIMARY KEY NOT NULL, 'name' TEXT NOT NULL)''')
bdd.execute('''CREATE TABLE IF NOT EXISTS 'Events' ('eid' TEXT PRIMARY KEY NOT NULL, 'etype' TEXT NOT NULL, 'init' TEXT NOT NULL, 'msg' TEXT NOT NULL, 'time' INTEGER NOT NULL)''')
lastreadMsg = {}
lastreadEvt = {}
Users = {}

start_server = websockets.serve(Loop, '0.0.0.0', 8765, ssl=sslcontext)

asyncio.get_event_loop().run_until_complete(start_server)
asyncio.get_event_loop().run_forever()